// to be done: boardIndex_

MixerChannel {
	// interface class for basic level and panning faders
	// each instance stores numchannels, level and panning info,
	// and pointers to the Synth object on the server
	// MixerChannel.mixer builds a gui for all active MixerChannels
	
	// recommended use: create MixerChannel first, then place playing
	// synths at the head of mixer_instance.group; effects can be added
	// at the head of the group too
	
	// this class uses my class BusDict to store names of active buses!
	// H. James Harkins - jamshark70@yahoo.com
	
	classvar	<servers;	// collection of active servers. see *new for why
					// dictionary of servers -> dictionary of bus numbers & mixerchannels
					// needed for autorepatching -- order of execution issues
	
	var	<name,			// mnemonic string, provided at creation
		<level, <pan,	// basic settings
		<server,			// the server
		<inbus, <outbus,	// buses: multiple MixerChannels can go to the same
						// outbus; these are not optional
		<inChannels, <outChannels,	// default is mono to stereo
		<postSendReady,	// if true, post sends are more efficient
						// should be false if post sends will not be used -- saves CPU
		<synth,			// synth object on server
//		<changroup,		// group for channel overall (contains fader & synthgroups)
		<fadergroup,			// group for faders/sends; contains synthgroup at head
		<synthgroup,			// group for things played on this channel
		<levelControl, <panControl,	// kr buses for level & pan
		<levelSynth, <panSynth,		// for automation
		<>postSendBus,	// for post-fader sends; prefader sends use inbus
		<>postSendXferSynth,	// see MixerPostSend for details
		<>preSends, <>postSends,	// keep track of sends for gui
		<>mcgui,					// set when MixerChannelGUI is created
		<>guiUpdateTime = 0.25,		// during LineTo, update every x sec's
		<levelguiUpdater, <panguiUpdater, 	// routines used in level/panLine
		patchesPlayed;		// so that free can tell crucial players to .stop
		
	*initClass {
		servers = Dictionary.new;
	}
	
///////////////////// CREATION AND INITIALIZATION /////////////////////

	*new { arg 	name = nil, server, inChannels = 1, outChannels = 2, 
				level = 0.75, pan = 0, postSendReady = false, inbus, outbus;

			// if this is the first mixerchannel on this server,
			// load all mixer and automation synthdefs
		servers.keys.includes(server).not.if({
			servers = servers.put(server, IdentityDictionary.new);
			server.sendMsg("/d_loadDir", "synthdefs/mixers");
			server.sendMsg("/d_loadDir", "synthdefs/mxaut");
		});
		
		^super.newCopyArgs(name, level, pan, server, 
			inbus, outbus, inChannels, outChannels, postSendReady).fixParms;
	}
	
	fixParms {	// make all parameters consistent; allows for
				// missing arguments
				// also adds mixer synth object to server in a new group
				// default outbus is standard output
		
		var synthdefName, argOutbus;
		
		argOutbus = outbus;	// save to determine if order-of-execution will have to be
							// adjusted

		preSends = Array.new;	// initialize these arrays
		postSends = Array.new;
		patchesPlayed = Array.new;

		inbus.isNil.if({	// if in/out buses are nil, create them
//("creating inbus").postln;
			inbus = BusDict.audio(server, 
					postSendReady.if({ outChannels }, { inChannels }), name ++ " in");
		});
		outbus.isNil.if({
//"creating outbus".postln;
			outbus = Bus.new(\audio, 0, outChannels, server);
		});
		
			// allow bus number as argument
			// note, if you supply a number for an
			// unallocated bus, you might not get an
			// error but your mixerchannel won't work
		inbus = inbus.isKindOf(Bus).if(
			{ inbus },	// if it's already a bus, leave it alone
			{ Bus.new(\audio, inbus, postSendReady.if({ outChannels }, { inChannels }), server) }
		);
		
		outbus.isKindOf(MixerChannel).if({
			outbus = outbus.inbus;		// route output of this to input of dest mc
		});
		
		outbus.isKindOf(Integer).if(
			{ outbus = Bus.new(\audio, outbus, outChannels, server);
			  argOutbus = outbus },		// integer will not be ok for order-of-exec test below
			{ nil }
		);
		
		name = (name.isNil).if({  // if name not given, lookup in BusDict
//("looking up name " ++ inbus.index).postln;
			name = BusDict.audioNames.at(server).at(inbus.index);
			(name == "" || name.isNil).if({	// still none,
				name = "Mixer " ++ inbus.index;	// name after bus number
			})
		}, { name }
		);

			// make group and drop synth onto server
//"creating group".postln;
			// since fadergroup contains synthgroup, order of execution of mixerchannels
			// can be changed by moving fadergroup
		fadergroup = Group.new(server);
		{ 	synthgroup = Group.head(fadergroup);
			nil
		}.defer(server.latency + 0.1);
			// SynthDefs for mixers must be named:
			// Mix[inChannels]x[outChannels]
			// eg: Mix1x2 is mono-to-stereo
		synthdefName = postSendReady.if({ "mixers/Mxb" }, { "mixers/Mix" })
			++ (inChannels.asString) ++ "x" ++ (outChannels.asString);
		{	// defer: we must allow time for the groupnode to become active
//("creating synth " ++ synthdefName).postln;
			synth = Synth.tail(fadergroup, synthdefName,
				[\busin, inbus.index, \busout, outbus.index,
				 \level, level, \pan, pan]);
				 // if outbus was specified, make sure dest comes after this
			argOutbus.notNil.if({ this.outbus_(outbus.asBus, false) });
			nil	// {} must return nil to execute only once
		}.defer(server.latency*2 + 0.2);
		
		servers.at(server).put(inbus.index, this);	// save for repatching purposes
			 
		^this
	}
	
	free { 
		arg updateGUI = true;	// by default, fix gui
							// but MixingBoard-free calls
							// this and shouldn't update
		(updateGUI && mcgui.isNil.not).if({	// if it's in a gui board,
				// pull from gui
				// false = removeAt shouldn't free mixer
				// since it's being freed here
			mcgui.board.remove(this, false);
		});
			// remove from mc dictionary
		servers.at(server).removeAt(inbus.index);
		this.freePatches; 		// stop all processes -- Patches first
//		preSends.do({ arg p; p.free });
//		postSends.do({ arg p; p.free });
//		synthgroup.free;
		fadergroup.free;
		BusDict.free(inbus);	// free buses
			// if outbus belongs to a mixer, don't free it
		outbus.asMixer.isNil.if({ BusDict.free(outbus); });
		levelControl.isNil.not.if({ BusDict.free(levelControl); });
		panControl.isNil.not.if({ BusDict.free(panControl); });
	}
	
	freePatches {
		patchesPlayed.do({ arg p; p.stop; });		// stop all processes
		patchesPlayed = Array.new;
	}		

	asBus {	// tell the outside world where I live
		^inbus
	}
	
	outbus_ { arg bus, updateGUI = true, target, moveAction;
			// repatch this channel to a new destination
			// moveAction should be \moveBefore, \moveAfter, \moveToHead, \moveToTail
		
		var mc;
		
		bus.isKindOf(Integer).if({	// allow integer in
			bus = Bus.new(\audio, bus, outChannels, server);
		});
		
		bus.isKindOf(MixerChannel).if({	// allow mixerchannel
			bus = bus.inbus;	// now isolate the bus for the following
		});
		
			// if the new bus has an entry in the MixerChannel table
		(mc = MixerChannel.servers.at(server).at(bus.index)).notNil.if({
				// if the dest mixer is routed to another mixer,
				// place the dest mixer before its destination
			(moveAction.isNil && mc.outbus.asMixer.notNil).if({
				moveAction = \moveBefore;
				target = mc.outbus.asMixer.fadergroup;
			}, {		// otherwise, if target not specified, move to tail of whole server
				moveAction = \moveToTail;
				target = target ? mc.server.asGroup;
			});
			mc.fadergroup.perform(moveAction, target);
		});

		outbus = bus;		// save
		synth.set(\busout, outbus.index);	// repatch
		updateGUI.if({ this.mcgui.refresh });
	}
	
///////////////////// INSTANT SETTING OF BASIC PARMS /////////////////////

	level_ { arg lev, updateGUI = true, stopAutomation = true;
			// flags are because a standard level setting by
			// the user should update gui and stop automation
			// but, other processes use this method that
			// should skip one or both of these

//("level_(" ++ lev ++ ", " ++ updateGUI ++ ", " ++ stopAutomation ++ ")").postln;

		level = lev;
		
			// deactivate automation if running
		stopAutomation.if({ this.stopLevelAuto });
		
		synth.set(\level, lev);		// plug in the value
		
		(updateGUI && mcgui.isNil.not).if({	// update GUI if it exists
				// use setProperty to prevent slider action from firing
			{ mcgui.levelUpdate(lev);
			  nil
			}.defer;	// so this can be done in a SystemClock routine
		});
	}
	
	pan_ { arg p, updateGUI = true, stopAutomation = true;

		pan = p;

		stopAutomation.if({ this.stopPanAuto 	});
		
		synth.set(\pan, p);		// plug in the value

		(updateGUI && mcgui.isNil.not).if({	// update GUI if it exists
			{ mcgui.panUpdate(p);
			  nil
			}.defer;	// so this can be done in a SystemClock routine
		});
	}
	
	name_ { arg n;	// give mixerchannel a new name and update gui

		name = n;
		
		mcgui.isNil.not.if({		// if it's in a board
			mcgui.nameUpdate(n);
		});
		^this
	}
			
	
///////////////////// BASIC AUTOMATION /////////////////////

	levelAuto {
			// place a .kr synthdef onto level control bus
		arg synthdef ... args;
		var dur, end;
		
			// if levelSynth is active, kill it
		this.stopLevelAuto;

			// schedule monitoring for gui

		args.includes(\end).if({	// if a finite synthdef,
			dur = args.at(args.indexOf(\dur) + 1);
			end = args.at(args.indexOf(\end) + 1);
			SystemClock.sched(		// schedule update for level inst var
				dur,	{ this.level_(end) }
			);
		});

			// create kr bus if it doesn't exist
		levelControl = levelControl ?? 
			BusDict.control(server, 1, name ++ " level");

//("Synth args: " ++ synthdef ++ ", ").post;
		levelSynth = Synth.head(fadergroup, "mxaut/" ++ synthdef, 
//(
			args.flat ++ args.includes(\busout).if(
				{ nil }, { [\busout, levelControl.index] })
//).postln;
		);
		synth.map(\level, levelControl.index);
		^levelSynth
	}
	
	panAuto {
			// .kr synthdef -> pan control bus
		arg synthdef ... args;
		var dur, end;
		
			// if panSynth is active, kill it
		this.stopPanAuto;

			// schedule monitoring for gui

		args.includes(\end).if({	// if a finite synthdef,
			dur = args.at(args.indexOf(\dur) + 1);
			end = args.at(args.indexOf(\end) + 1);
			SystemClock.sched(		// schedule update for pan inst var
				dur, { this.pan_(end) }
			);
		});

			// make kr bus if needed
		panControl = panControl ?? 
			BusDict.control(server, 1, name ++ " pan");

//("Synth args: " ++ synthdef ++ ", ").post;
		panSynth = Synth.head(fadergroup, "mixers2/" ++ synthdef, 
//(
			args.flat ++ args.includes(\busout).if(
				{ nil }, { [\busout, panControl.index] } )
//).postln
		);
		synth.map(\pan, panControl.index);
		^panSynth
	}
	
		// line automation -- ...LineTo uses current value as start
		// any warp that can be used in a ControlSpec can be used here
	levelLine {
		arg start = 0, end = 1, dur = 1, warp = \linear;
		
		var startTime, timeSpec;
		
		this.stopLevelAuto;	// begin by clearing automation

			// start the automation process on the server
			// this line will be revived when bus.get is implemented
//		this.levelAuto("krLine", \start, start, \end, end, \dur, dur);
		
			// schedule periodic updates for level and gui
			// spec to convert time into level
		timeSpec = ControlSpec.new(start, end, warp);
		startTime = Main.elapsedTime;
//"starting gui updater routine".postln;
		levelguiUpdater = Routine({
			{ (Main.elapsedTime - startTime) <= dur }.while({
				this.level_(timeSpec.map(
					(Main.elapsedTime-startTime) / dur
				), true, false);  // update gui but don't stop automation
				guiUpdateTime.yield	// wait before going again
			});
//"stopping routine".postln;
			nil.yield		// stop
		}).play(SystemClock);
	}
	
	levelTo {		// start from current level
		arg end, dur, warp = \linear;
		
		this.levelLine(level, end, dur, warp);
	}

	panLine {
		arg start = 0, end = 0, dur = 1, warp = \linear;
		
		var startTime, timeSpec;

		this.stopPanAuto;
		
			// start the automation process on the server
//		this.panAuto("krLine", \start, start, \end, end, \dur, dur);
		
			// schedule periodic updates for level and gui
			// spec to convert time into level
		timeSpec = ControlSpec.new(start, end, warp);
		startTime = Main.elapsedTime;

		panguiUpdater = Routine({
			{ (Main.elapsedTime - startTime) <= dur }.while({
				this.pan_(timeSpec.map(
					(Main.elapsedTime-startTime) / dur
				), true, false);
				guiUpdateTime.yield	// wait before going again
			});
			nil.yield		// stop
		}).play(SystemClock);
	}
	
	panTo {		// start from current level
		arg end, dur, warp = \linear;
		
		this.panLine(pan, end, dur, warp);
	}
	
	stopLevelAuto {
		levelguiUpdater.isNil.not.if({ 
			levelguiUpdater.stop;
			levelguiUpdater = nil;
		});
		levelSynth.notNil.if({ 
			levelSynth.free;	// kill the synth
			levelSynth = nil;
			synth.map(\level, -1);	// remove the mapping but keep the bus (might use later)
		});
		^this
	}
	
	stopPanAuto {
		panguiUpdater.isNil.not.if({ 
			panguiUpdater.stop;
			panguiUpdater = nil;
		});
		panSynth.notNil.if({ 
			panSynth.free;
			panSynth = nil;
			synth.map(\pan, -1);
		});
		^this
	}
		
///////////////////// BASIC PLAYING /////////////////////

	play { arg thing ... args;
			// input is string, treat as synthdef name
		thing.isKindOf(String).if({
			args = args.flat;
				// if outbus is not included in parm list,
				// add it now
			args = args ++ args.includes(\outbus).if({ nil }, { [\outbus, inbus.index] });
//(thing ++ ": ").post; args.postln;
			^Synth.tail(synthgroup, thing, args);
		});
		
		thing.isKindOf(Instr).if({	// wrap Instr's inside a Patch
			thing = Patch(thing, args.flat);
		});
		
			// play a Patch on my inbus
		(thing.class.superclass.name == \AbstractPlayer).if({			thing.play(synthgroup, 0, inbus);
			patchesPlayed = patchesPlayed.add(thing);
			^thing	// this way, user can say myPatch = m.play(Instr(...));
		});

//			// is there a way to plug in outbus automatically?
//		thing.isKindOf(Routine).if({
//			SystemClock.play(thing);
//		});
	}
	
	playfx { arg thing ... args;
		thing.isKindOf(String).if({
			args = args.flat;
				// if outbus is not included in parm list,
				// add it now
			args = args ++ args.includes(\outbus).if({ nil }, { [\outbus, inbus.index] });
//(thing ++ ": ").post; args.postln;
				// play at tail of synth group
			^Synth.tail(synthgroup, thing, args);
		});
	}

///////////////////// SENDS /////////////////////

	newPreSend { arg dest, level;		// bus number or mixerchannel
		preSends = preSends.add(MixerPreSend.new(this, dest, level));
		mcgui.notNil.if({ mcgui.refresh });
	}
	
	newPostSend { arg dest, level;
		postSends = postSends.add(MixerPostSend.new(this, dest, level));
		mcgui.notNil.if({ mcgui.refresh });
	}

}