MixingBoard {
	// group a bunch of MixerChannels into a console
	// H. James Harkins - jamshark70@yahoo.com
	
	classvar	<>defaultSkin,	// override MixerSkin's default
			<boards;			// collection of boards
							// so MixerChannel can update GUI
							
	var	<mixers,	// array of active MixerChannelGUI objects
		<w,		// gui window
		<skin,	// customize display
		<>name;	// window title bar
		
	*initClass {
		Class.initClassTree(MixerSkin);
		defaultSkin = MixerSkin.new;
	}
	
	*new { arg name = "Mixer", skin ... mixers;
		^super.new.init(name, skin, mixers);
	}
	
	init { arg name1, skin1 ... mixers1;

		mixers.do({ arg m; m.free; });	// free all mixers if they exist
		
		mixers = Array.new;

		name = name1;
		skin = skin1 ?? defaultSkin.copy;	// if no skin specified, use default
										// use copy so default is untouched
		
		boards = boards.add(this);	// save in collection
		^this.buildGUI(mixers1);	// draw it
	}
	
	free {
		// remove this mixing board and all channels
		
			// false says to hold off gui updates
		mixers.do({ arg m; m.mixer.free(false); });
		w.close;
		boards.remove(this);
//		MixingBoard.renumberChannels;	// fix indices in MixerChannels
	}
	
	add { arg ... mxs;
		// add mixers to console
		var oldsize;
		oldsize = mixers.size;	// save old size so we know where to
							// start making new views
		mxs = mxs.flat;
		mixers = mixers ++ mxs.collect({ 
			arg m, i;
			MixerChannelGUI.new(m, this, i + oldsize);
		});
		
			// fix window, draw controls, fix numbering and redraw
		this.sizeWindow;
		{ this.refresh; nil }.defer(0.1);
		^this
	}
	
		// remove & removeAt won't work yet
	remove { arg ... mxs;
		// remove mixer(s) from console
		// mxs may be a single mixerchannel, collection or chan, chan, chan...
		var ind, mxtemp;

		mxs.isKindOf(MixerChannel).if({
			mxs = [ mxs ];	// turn into a collection
		}, {
			mxs = mxs.flat;	// else flatten
		});
		
			// extract mixerchannels from mcguis
		mxtemp = mixers.collect({ arg m; m.mixer });
		
			// get indices
		ind = mxs.collect({ arg m; 
			m.isKindOf(Boolean).if({ m },	// leave Booleans alone
				{ mxtemp.indexOf(m) });		// others, get index of m_chan
		});
		
		^this.removeAt(ind);
	}
	
	removeAt { arg ... ind;
		var freeIt;
		
		ind = ind.flat;
		
		freeIt = ind.last;		// last arg might be boolean
		freeIt.isKindOf(Boolean).if({
			ind.remove(freeIt);
		}, {
			freeIt = true;	// if not, assume yes, this should free the mc
		});
		
			// sort in descending order
			// descending b/c if we remove from the beginning,
			// later indexes will change and we'll remove the wrong channels
		
		ind = ind.sort({ arg a,b; a > b });
//ind.postln;
//freeIt.postln;		

		ind.do({ arg i;	// loop over indexes
				// clear associated views
				// also frees mixerchannel if freeIt == true
			mixers.at(i).free(freeIt);
			
			mixers.removeAt(i);	// delete from collection
			
		});

		this.sizeWindow;	// resize window
		{ this.refresh; nil }.defer(0.1);	// redraw after size change takes

		^this
	
	}

		// GUI methods
	buildGUI {		// open a window for all channels
		arg mx;		// channels to add
		var newWindow = false;	// are we creating this new?

			// make window first		
		{ // for defer
		
		w.isNil.if({		// only build it if no window exists yet
			w = SCWindow(name);
			w.onClose_({ this.free });	// if user closes window, clean up after yourself
			newWindow = true;
		});

		mx = mx.flat;	// allows ... passing from higher levels

		mx.do({ arg mx, i;
			mixers = mixers.add(MixerChannelGUI.new(mx, this, i));
		});
		
		this.sizeWindow(newWindow);

			// must defer refresh of controls
			// because it takes time for w.bounds to
			// catch up with w.bounds_
		{ 	this.refresh; 
			newWindow.if({ w.front });
			nil
		}.defer(0.25);
		
		nil }.defer;
	}
	
	sizeWindow {
		arg isNew = false;		// set by buildGUI
		var num, hSize, vSize;
		var rows, columns;

		num = mixers.size;
			// get size of window
			// 1st, assume everything is horizontal
		rows = 1; columns = num;
			// if too many to fit on one row
		((num * (skin.channelSize.x + skin.gap.x)
				+ skin.gap.x) > skin.maxSize.x).if({  // maximum columns
			columns = ((skin.maxSize.x-skin.gap.x) /
				(skin.channelSize.x+skin.gap.x)).asInteger;
				// note, it's possible to have more rows than will fit on screen
			rows = (num / columns).asInteger;	// make enough rows
			(num > (columns*rows)).if({		// if a remainder
				rows = rows+1;
			});
		});
		
		hSize = (columns * (skin.channelSize.x+skin.gap.x)) + skin.gap.x;
		vSize = (rows * (skin.channelSize.y+skin.gap.y))
				+ skin.gap.y + 20;		// another 20 for window header
		
		{
		isNew.if({	// if new, assign origin
			w.bounds = Rect.new(5, 5, hSize, vSize);
		}, {			// if old, keep existing origin
			w.bounds = Rect.new(w.bounds.left, w.bounds.top, hSize, vSize);
		});
		nil }.defer;
		
		^this
	}		
	
	refresh {
		var	origin,	// of this mixerchannel
			sendItems,	// to build dest menu
			send;		// to fill in send menu values
			
			// start here for first channel's group of controls
		origin = Point(skin.gap.x, skin.gap.y);

			// for each mixerchannel
		mixers.do({ arg m, i;
			m.origin = origin;		// reset origin
			m.refresh;		// redraw

				// move to next origin
			origin = origin + Point(skin.channelSize.x + skin.gap.x);

				// within horizontal bounds?
				// if not, wrap around
			((origin.x + skin.channelSize.x) > w.bounds.width).if({
				origin = Point(skin.gap.x,
					origin.y + skin.channelSize.y + skin.gap.y);
			});
		});
		^this
	}
	
//	renumberChannels {
//			// if mixers or boards collections change, the associated
//			// mixers have to be renumbered and the actions fixed
//		var board;	// index to board
//		board = boards.indexOf(this);
//		mixers.do({ arg m, i;
//			m.boardIndex = board;
//			m.mixerIndex = i;
//			
//				// refresh actions, since they refer to specific indices
//			levelSliders.at(i).action = this.levelAction(i);
//			panSliders.at(i).action = this.panAction(i);
//			
//			skin.preSendsShow.do({ arg j;
//				// TODO: adjust for 1st send displayed
//				((m.preSends)@j).notNil.if({ ((m.preSends)@j).guiIndex = j; });
//				(preSendMenus@i@j).action = this.preSendMenuAction(i, j);
//				(preSendLevels@i@j).action = this.preSendLevelAction(i, j);
//			});
//			skin.postSendsShow.do({ arg j;
//				// TODO: adjust for 1st send displayed
//				((m.postSends)@j).notNil.if({ ((m.postSends)@j).guiIndex = j; });
//				(postSendMenus@i@j).action = this.postSendMenuAction(i, j);
//				(postSendLevels@i@j).action = this.postSendLevelAction(i, j);
//			});
//		});
//		^this
//	}
//	
//	*renumberChannels {
//			// do it for every active board
//		boards.do({ arg b; b.renumberChannels });
//	}
	

	// change appearance
	// to change one parameter, if b is the board, do
	// b.skin = b.skin.color1_(Color.rgb(....)); 
	skin_ { arg newskin;
		skin = newskin;
		this.sizeWindow;
		{ this.refresh; nil }.defer(0.1);	// redraw gui
		^this
	}
	
	*facelift { 
		// refresh all boards
		boards.do({ arg b;
			b.sizeWindow;
			{ b.refresh; nil }.defer(0.1);	// defer allows resize to take
		});
	}
}